
from .request import Request


def verify(
        receipt_data, password=None, exclude_old_transactions=False, **kwargs):
    """Shortcut API for :class:`itunesiap.request.Request`

    :param str receipt_data: An iTunes receipt data as Base64 encoded string.
    :param proxy_url: A proxy url to access the iTunes validation url
    :param bool use_production: Override environment value if given
    :param bool use_sandbox: Override environment value if given
    :param bool verify_ssl: Override environment value if given

    :return: :class:`itunesiap.receipt.Receipt` object if succeed.
    :raises: Otherwise raise a request exception.
    """
    proxy_url = kwargs.pop('proxy_url', None)
    request = Request(
        receipt_data, password, exclude_old_transactions, proxy_url=proxy_url)
    return request.verify(**kwargs)




import json
import functools

import requests

from . import receipt
from . import exceptions
from .environment import Environment

RECEIPT_PRODUCTION_VALIDATION_URL = "https://buy.itunes.apple.com/verifyReceipt"
RECEIPT_SANDBOX_VALIDATION_URL = "https://sandbox.itunes.apple.com/verifyReceipt"
STATUS_SANDBOX_RECEIPT_ERROR = 21007


class Request(object):
    """Validation request with raw receipt.

    Use `verify` method to try verification and get Receipt or exception.

    :param str receipt_data: An iTunes receipt data as Base64 encoded string.
    :param proxy_url: A proxy url to access the iTunes validation url.
    """

    def __init__(
            self, receipt_data, password=None, exclude_old_transactions=False,
            **kwargs):
        self.receipt_data = receipt_data
        self.password = password
        self.exclude_old_transactions = exclude_old_transactions
        self.proxy_url = kwargs.pop('proxy_url', None)

    def __repr__(self):
        return u'<Request({0}...)>'.format(self.receipt_data[:20])

    @property
    def request_content(self):
        """Build request body for iTunes."""
        request_content = {
            'receipt-data': self.receipt_data,
            'exclude-old-transactions': self.exclude_old_transactions}
        if self.password is not None:
            request_content['password'] = self.password
        return request_content

    def verify_from(self, url, verify_ssl=True):
        """Try verification from given url.

        :param str url: iTunes verification API URL.
        :param bool verify_ssl: SSL verification.

        :return: :class:`itunesiap.receipt.Receipt` object if succeed.
        :raises: Otherwise raise a request exception.
        """
        # If the password exists from kwargs, pass it up with the request, otherwise leave it alone
        post_body = json.dumps(self.request_content)
        if self.proxy_url:
            protocol = self.proxy_url.split('://')[0]
            requests_post = functools.partial(requests.post, proxies={protocol: self.proxy_url})
        else:
            requests_post = requests.post
        try:
            http_response = requests_post(url, post_body, verify=verify_ssl)
        except requests.exceptions.RequestException as e:
            raise exceptions.ItunesServerNotReachable(exc=e)

        if http_response.status_code != 200:
            raise exceptions.ItunesServerNotAvailable(http_response.status_code, http_response.content)

        response = receipt.Response(json.loads(http_response.content.decode('utf-8')))
        if response.status != 0:
            raise exceptions.InvalidReceipt(response.status, response=response)
        return response

    def verify(self, **options):
        """Try verification with current environment.
        If verify_ssl is true, Apple's SSL certificiate will be
        verified. The verify_ssl is set to false by default for
        backwards compatibility.

        :param Environment env: Override environment if given
        :param bool use_production: Override environment value if given
        :param bool use_sandbox: Override environment value if given
        :param bool verify_ssl: Override environment value if given

        :return: :class:`itunesiap.receipt.Receipt` object if succeed.
        :raises: Otherwise raise a request exception.
        """
        env = options.get('env')
        if not env:
            env = Environment.current()
        use_production = options.get('use_production', env.use_production)
        use_sandbox = options.get('use_sandbox', env.use_sandbox)
        verify_ssl = options.get('verify_ssl', env.verify_ssl)
        assert(env.use_production or env.use_sandbox)

        response = None
        if use_production:
            try:
                response = self.verify_from(RECEIPT_PRODUCTION_VALIDATION_URL, verify_ssl)
            except exceptions.InvalidReceipt as e:
                if not use_sandbox or e.status != STATUS_SANDBOX_RECEIPT_ERROR:
                    raise

        if not response and use_sandbox:
            try:
                response = self.verify_from(RECEIPT_SANDBOX_VALIDATION_URL, verify_ssl)
            except exceptions.InvalidReceipt:
                raise

        return response




from prettyexc import PrettyException as E


class RequestError(E):
    pass


class ItunesServerNotAvailable(RequestError):
    pass


class ItunesServerNotReachable(ItunesServerNotAvailable):
    pass


class InvalidReceipt(RequestError):
    _req_kwargs_keys = ['status']
    _descriptions = {
        21000: 'The App Store could not read the JSON object you provided.',
        21002: 'The data in the receipt-data property was malformed.',
        21003: 'The receipt could not be authenticated.',
        21004: 'The shared secret you provided does not match the shared secret on file for your account.',
        21005: 'The receipt server is not currently available.',
        21006: 'This receipt is valid but the subscription has expired. When this status code is returned to your server, the receipt data is also decoded and returned as part of the response.',
        21007: 'This receipt is a sandbox receipt, but it was sent to the production service for verification.',
        21008: 'This receipt is a production receipt, but it was sent to the sandbox service for verification.',
    }

    @property
    def description(self):
        return self._descriptions.get(self.status, None)


class MissingFieldError(E, AttributeError, KeyError):
    pass



"""
itunes-iap
~~~~~~~~~~

Itunes In-app Purchase verification api.

:copyright: (c) 2013 Jeong YunWon - 2014 Andy Briggs
:license: 2-clause BSD.
"""

from six import PY3

from .request import Request
from .receipt import Response, Receipt, InApp
from .shortcut import verify

from . import exceptions
from . import environment

exc = exceptions
env = environment  # env.default, env.sandbox, env.review

try:
    import pkg_resources
except ImportError:  # pragma: no cover
    __version__ = VERSION = None
else:
    __version__ = pkg_resources.resource_string(
        'itunesiap', 'version.txt').strip()
    if PY3:
        __version__ = __version__.decode('ascii')  # pragma: no cover
    VERSION = tuple(int(v) for v in __version__.split('.'))


__all__ = (
    'Request', 'Response', 'Receipt', 'InApp', 'verify',
    'exceptions', 'exc', 'environment', 'env')




import pytz
import dateutil.parser
import warnings
import json

from .exceptions import MissingFieldError
from .tools import lazy_property


def _to_datetime(value):
    """Try to parse Apple iTunes receipt date format.

    By reference, they insists it is rfc3339:

        https://developer.apple.com/library/ios/releasenotes/General/ValidateAppStoreReceipt/Chapters/ReceiptFields.html#//apple_ref/doc/uid/TP40010573-CH106-SW4

    Though data I got from apple server does not. So the strategy is:

        - Give them a chance anyway.
        - Or split timezone string and read it from pytz.
    """
    try:
        d = dateutil.parser.parse(value)
    except ValueError as e:
        value, timezone = value.rsplit(' ', 1)
        try:
            d = dateutil.parser.parse(value + '+00:00')
        except ValueError:
            raise e
        d = d.replace(tzinfo=pytz.timezone(timezone))
    return d


def _to_bool(data):
    assert data in ('true', 'false'), \
        ("Cannot convert {0}, "
         "acceptable values are true' and 'false'".format(data))
    return json.loads(data)


class ObjectMapper(object):
    """Pretty interface for decoded receipt object.
    """
    __WHITELIST__ = []
    __EXPORT_FILTERS__ = {}

    def __init__(self, data):
        self._ = data

    def __repr__(self):
        return u'<{0}({1})>'.format(self.__class__.__name__, self._)

    def __getitem__(self, item):
        return self._[item]

    def __getattr__(self, item):
        try:
            return super(ObjectMapper, self).__getattr__(item)
        except AttributeError:
            try:
                if item.startswith('_'):
                    key = item[1:]
                    if key not in self.__WHITELIST__:
                        warnings.warn('Given key `{0}` is not in __WHITELIST__. It maybe a wrong key. Check raw data `_` for real receipt data.'.format(key))
                    return self._[key]
                if item in self.__EXPORT_FILTERS__:
                    filter = self.__EXPORT_FILTERS__[item]
                    return filter(self._[item])
                if item in self.__WHITELIST__:
                    return self._[item]
            except KeyError:
                raise MissingFieldError(item)
            return super(ObjectMapper, self).__getattribute__(item)


class Response(ObjectMapper):
    """The root response.

    status: See https://developer.apple.com/library/ios/releasenotes/General/ValidateAppStoreReceipt/Chapters/ValidateRemotely.html#//apple_ref/doc/uid/TP40010573-CH104-SW1
    """
    __WHITELIST__ = ['receipt', 'latest_receipt']  # latest_receipt_info
    __EXPORT_FILTERS__ = {'status': int}

    @lazy_property
    def receipt(self):
        return Receipt(self._receipt)

    @lazy_property
    def latest_receipt(self):
        return Receipt(self._latest_receipt)


class Receipt(ObjectMapper):
    __WHITELIST__ = ['in_app']
    __EXPORT_FILTERS__ = {}

    @lazy_property
    def in_app(self):
        if 'in_app' in self._:
            return list(map(InApp, self._in_app))
        else:
            return [InApp(self._)]

    @property
    def last_in_app(self):
        return self.in_app[-1]


class InApp(ObjectMapper):
    __WHITELIST__ = [
        'quantity', 'product_id', 'transaction_id', 'original_transaction_id', 'is_trial_period',
        'purchase_date', 'original_purchase_date', 'expires_date', 'cancellation_date',
        'purchase_date_ms', 'original_purchase_date_ms', 'expires_date_ms', 'cancellation_date_ms', 'expires_date_formatted', ]
    __EXPORT_FILTERS__ = {
        'quantity': int,
        'is_trial_period': _to_bool,
        'purchase_date': _to_datetime,
        'original_purchase_date': _to_datetime,
        'expires_date': _to_datetime,
        'expires_date_formatted': _to_datetime,
        'cancellation_date': _to_datetime,
        'purchase_date_ms': int,
        'original_purchase_date_ms': int,
        'expires_date_ms': int,
        'cancellation_date_ms': int,
    }




from . import exceptions
from .tools import deprecated
import requests
import json
import contextlib
from prettyexc import PrettyException as E


RECEIPT_PRODUCTION_VALIDATION_URL = "https://buy.itunes.apple.com/verifyReceipt"
RECEIPT_SANDBOX_VALIDATION_URL = "https://sandbox.itunes.apple.com/verifyReceipt"


USE_PRODUCTION = True
USE_SANDBOX = False


class ModeNotAvailable(E):
    message = '`mode` should be one of `production`, `sandbox`, `review`, `reject`'


exceptions.ModeNotAvailable = ModeNotAvailable


def config_from_mode(mode):
    if mode not in ('production', 'sandbox', 'review', 'reject'):
        raise exceptions.ModeNotAvailable(mode)
    production = mode in ('production', 'review')
    sandbox = mode in ('sandbox', 'review')
    return production, sandbox


def set_verification_mode(mode):
    """Set global verification mode that where allows production or sandbox.
    `production`, `sandbox`, `review` or `reject` available. Otherwise raise
    an exception.

    `production`: Allows production receipts only. Default.
    `sandbox`: Allows sandbox receipts only.
    `review`: Allows production receipts but use sandbox as fallback.
    `reject`: Reject all receipts.
    """
    global USE_PRODUCTION, USE_SANDBOX
    USE_PRODUCTION, USE_SANDBOX = config_from_mode(mode)


def get_verification_mode():
    if USE_PRODUCTION and USE_SANDBOX:
        return 'review'
    if USE_PRODUCTION:
        return 'production'
    if USE_SANDBOX:
        return 'sandbox'
    return 'reject'


class Request(object):
    """Validation request with raw receipt. Receipt must be base64 encoded string.

    Use `verify` method to try verification and get Receipt or exception.
    """

    def __init__(self, receipt, password=None, **kwargs):
        self.receipt = receipt
        self.password = password
        self.use_production = kwargs.get('use_production', USE_PRODUCTION)
        self.use_sandbox = kwargs.get('use_sandbox', USE_SANDBOX)
        self.verify_ssl = kwargs.get('verify_ssl', False)
        self.response = None

    def __repr__(self):
        valid = None
        if self.result:
            valid = self.result['status'] == 0
        return u'<Request(valid:{0}, data:{1}...)>'.format(valid, self.receipt[:20])

    @property
    def result(self):
        return self._extract_receipt(json.loads(self.response.content.decode('utf-8')))

    @property
    def request_content(self):
        if self.password is not None:
            request_content = {'receipt-data': self.receipt, 'password': self.password}
        else:
            request_content = {'receipt-data': self.receipt}
        return request_content

    def verify_from(self, url, verify_ssl=False):
        """Try verification from given url."""
        # If the password exists from kwargs, pass it up with the request, otherwise leave it alone
        post_body = json.dumps(self.request_content)
        try:
            self.response = requests.post(url, post_body, verify=verify_ssl)
        except requests.exceptions.RequestException:
            raise

        if self.response.status_code != 200:
            raise exceptions.ItunesServerNotAvailable(self.response.status_code, self.response.content)

        status = self.result['status']
        if status != 0:
            raise exceptions.InvalidReceipt(status, receipt=self.result.get('receipt', None))
        return self.result

    def _extract_receipt(self, receipt_data):
        """There are two formats that iTunes iap purchase receipts are
        sent back in
        """
        if 'receipt' not in receipt_data:
            return receipt_data
        in_app_purchase = receipt_data['receipt'].get('in_app', [])
        if len(in_app_purchase) > 0:
            receipt_data['receipt'].update(in_app_purchase[-1])
        return receipt_data

    @deprecated
    def validate(self):
        return self.verify()

    def verify(self, verify_ssl=None):
        """Try verification with settings. Returns a Receipt object if succeeded.
        Or raise an exception. See `self.response` or `self.result` to see details.
        """
        assert(self.use_production or self.use_sandbox)
        if verify_ssl is None:
            verify_ssl = self.verify_ssl
        if self.use_production:
            try:
                receipt = self.verify_from(RECEIPT_PRODUCTION_VALIDATION_URL, verify_ssl)
            except exceptions.InvalidReceipt:
                if not self.use_sandbox:
                    raise

        if self.use_sandbox:
            try:
                receipt = self.verify_from(RECEIPT_SANDBOX_VALIDATION_URL, verify_ssl)
            except exceptions.InvalidReceipt:
                raise

        return Receipt(receipt)

    @contextlib.contextmanager
    def verification_mode(self, mode):
        configs = self.use_production, self.use_sandbox
        self.use_production, self.use_sandbox = config_from_mode(mode)
        yield
        self.use_production, self.use_sandbox = configs


class Receipt(object):
    """Pretty interface for decoded receipt object.
    """
    def __init__(self, data):
        self.data = data
        self.receipt = data['receipt']
        self.receipt_keys = list(self.receipt.keys())

    def __repr__(self):
        return u'<Receipt({0}, {1})>'.format(self.status, self.receipt)

    @property
    def _(self):
        return self.data

    @property
    def status(self):
        return self.data['status']

    @property
    def latest_receipt(self):
        return self.data['latest_receipt']

    def __getattr__(self, key):
        if key in self.receipt_keys:
            return self.receipt[key]
        try:
            return super(Receipt, self).__getattr__(key)
        except AttributeError:
            return super(Receipt, self).__getattribute__(key)


def verify(data, test_paid=lambda id: id):
    """Convenient verification shortcut.

    :param data: iTunes receipt data
    :param test_paid: Function to test the receipt is paid. Function should
        raise error to disallow response. Parameter is `original_transaction_id`
    :return: :class:`itunesiap.core.Response`
    """
    request = Request(data)
    response = request.verify()
    test_paid(response.original_transaction_id)
    return response
